﻿using System.Collections.Generic;
using System.Reflection.Emit;

namespace MyMapper
{
    /// <summary>
    /// Mapper implementation using LCG (Lightweight Code Generation). 
    /// </summary>
    public class Mapper : MapperBase 
    {
        private delegate void CopyPublicPropertiesDelegate<T, TU>(T source, TU target);
        private readonly Dictionary<string, object> _del = new Dictionary<string, object>();

        /// <summary>
        /// Maps the given types using given property map. This method can be used to provide 
        /// mapper with custom property map.
        /// </summary>
        /// <typeparam name="T">Source object where property values are copied from.</typeparam>
        /// <typeparam name="TU">Target object where property values are copied to.</typeparam>
        /// <param name="maps">Custom property map for given two types.</param>
        public override void MapTypes<T, TU>(IList<PropertyMap> maps)
        {
            var key = GetMapKey<T, TU>();

            var source = typeof(T);
            var target = typeof(TU);

            var args = new[] { source, target };
            var mod = typeof(Mapper).Module;

            var dm = new DynamicMethod(key, null, args, mod);
            var il = dm.GetILGenerator();            

            foreach (var map in maps)
            {
                il.Emit(OpCodes.Ldarg_1);
                il.Emit(OpCodes.Ldarg_0);
                il.EmitCall(OpCodes.Callvirt, map.SourceProperty.GetGetMethod(), null);
                il.EmitCall(OpCodes.Callvirt, map.TargetProperty.GetSetMethod(), null);
            }
            il.Emit(OpCodes.Ret);
            var del = dm.CreateDelegate(typeof(CopyPublicPropertiesDelegate<T, TU>));

            if (_del.ContainsKey(key))
                _del[key] = del;
            else 
                _del.Add(key, del);
        }

        /// <summary>
        /// Creates the mapping between given types. This method uses base class property 
        /// mapper to get property mapping.
        /// </summary>
        /// <typeparam name="T">Source type from where property values are copied.</typeparam>
        /// <typeparam name="TU">Target type to where property values are copied.</typeparam>
        public override void MapTypes<T, TU>()
        {
            var key = GetMapKey<T, TU>();
            if (_del.ContainsKey(key))
                return;

            var maps = GetMatchingProperties<T, TU>();
            MapTypes<T, TU>(maps);
        }

        /// <summary>
        /// Copies property values from source object to target object using previously created 
        /// mapping. Mapping must exist before copy operation is called.
        /// </summary>
        /// <typeparam name="T">Type of source object.</typeparam>
        /// <typeparam name="TU">Type of target object.</typeparam>
        /// <param name="source">Source object where property values are copied from.</param>
        /// <param name="target">Target object where property values are copied to.</param>
        public override void Copy<T, TU>(T source, TU target)
        {
            var key = GetMapKey<T, TU>();
            var del = (CopyPublicPropertiesDelegate<T, TU>)_del[key];
            del.Invoke(source, target);
        }
    }
}